use rltk::{field_of_view, Point, RandomNumberGenerator};
use specs::prelude::*;

use crate::{
    components::{Player, Position, Viewshed},
    map::Map,
    GameLog, Hidden, Name,
};

pub struct VisibilitySystem {}

impl<'a> System<'a> for VisibilitySystem {
    type SystemData = (
        WriteExpect<'a, Map>,
        Entities<'a>,
        WriteStorage<'a, Viewshed>,
        WriteStorage<'a, Position>,
        ReadStorage<'a, Player>,
        WriteStorage<'a, Hidden>,
        WriteExpect<'a, RandomNumberGenerator>,
        ReadStorage<'a, Name>,
        WriteExpect<'a, GameLog>,
    );

    fn run(&mut self, data: Self::SystemData) {
        let (
            mut map,
            entities,
            mut viewshed,
            pos,
            player,
            mut hidden,
            mut rng,
            names,
            mut game_log,
        ) = data;
        for (ent, viewshed, pos) in (&entities, &mut viewshed, &pos).join() {
            if !viewshed.dirty {
                continue;
            }

            viewshed.dirty = false;
            viewshed.visible_tiles = field_of_view(Point::new(pos.x, pos.y), viewshed.range, &*map);
            viewshed
                .visible_tiles
                .retain(|p| p.x >= 0 && p.x < map.width && p.y >= 0 && p.y < map.height);

            let p: Option<&Player> = player.get(ent);
            if let Some(_p) = p {
                for t in map.visible_tiles.iter_mut() {
                    *t = false
                }
                for vis in viewshed.visible_tiles.iter() {
                    let idx = map.xy_idx(vis.x, vis.y);
                    map.revealed_tiles[idx] = true;
                    map.visible_tiles[idx] = true;

                    for e in map.tile_content[idx].iter() {
                        if let Some(_maybe_hidden) = hidden.get(*e) {
                            if rng.roll_dice(1, 24) == 1 {
                                if let Some(name) = names.get(*e) {
                                    game_log
                                        .entries
                                        .push(format!("You spotted a {}.", &name.name));
                                }
                                hidden.remove(*e);
                            }
                        }
                    }
                }
            }
        }
    }
}
